/* $Id: sh3eva.c,v 1.5 1999/09/02 10:12:26 stuart Exp $
 *
 * sh3mon/sh3eva.c
 *
 * Copyright (C) 1999 Stuart Menefy
 */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "sh3mon.h"

static int fd;

static int local_open(const char* name)
{
  int fd;
  char buffer[100];
  struct termios term;
  int result;

  fd=open(name, O_RDWR | O_NDELAY | O_NOCTTY);
  if (fd < 0) {
    sprintf(buffer, "Failed to open %s", name);
    perror(buffer);
    exit(1);
  }

  result = tcgetattr(fd, &term);
  if (result != 0) {
    sprintf(buffer, "ioctl TCGETS failed on %s\n", name);
    perror(buffer);
    exit(1);
  }

  term.c_cc[VMIN]  = 1;
  term.c_cc[VTIME] = 0;
  term.c_lflag = 0;
  term.c_cflag = B38400 | CS8 | CREAD | CLOCAL;
  term.c_iflag = IGNBRK;
  term.c_oflag = 0;

  result = tcsetattr(fd, TCSAFLUSH, &term);
  if (result != 0) {
    sprintf(buffer, "ioctl TCSETS failed on %s\n", name);
    perror(buffer);
    exit(1);
  }

  result = fcntl(fd, F_GETFL);
  if (result == -1) {
    sprintf(buffer, "fcntl F_GETFL failed on %s\n", name);
    perror(buffer);
    exit(1);
  }

  result = fcntl(fd, F_SETFL, result & (~ O_NDELAY));
  if (result == -1) {
    sprintf(buffer, "fcntl F_SETFL failed on %s\n", name);
    perror(buffer);
    exit(1);
  }

  return fd;
}

static void unknown(unsigned char*, int);
static void memmap(unsigned char*, int);
static void regdump(unsigned char*, int);

typedef struct {
  int len;
  unsigned char data[2];
  void (*fn)(unsigned char*, int);
  int results;
} send_data_t;

#define SEND_DATA_LEN 7
send_data_t send_data[SEND_DATA_LEN] = {
  { 1, {0x03 },      unknown, 1 },
  { 2, {0x05, 0x00}, unknown, 2 },
  { 1, {0x00 },      unknown, 1 },
  { 1, {0x01 },      memmap,  1 },
  { 1, {0x01 },      memmap,  1 },
  { 1, {0x00 },      unknown, 1 },
  { 2, {0x05, 0x21}, regdump, 2 },
  /*  { 2, {0x00, 0x00}, unknown, 1 } */
};

static void
dump_buffer(const unsigned char* buffer, int count)
{
  int i, j;
  int offset;
  char c;

  for (offset = 0; offset < count; offset += 8) {
    j = count - offset;
    if (j > 8) j = 8;
    printf("%04x : ", offset);
    for (i=0; i<j; i++) {
      c=buffer[offset + i];
      printf("%02x ", c & 0xff);
    }
    for ( ; i<8; i++) {
      printf("   ");
    }
    for (i=0; i<j; i++) {
      c=buffer[offset + i];
      printf("%c", isprint(c) ? c : '.');
    }
    printf("\n");
  }
}

static int
read_result(int fd, unsigned char* buffer, size_t size)
{
  int count;
  enum { rrs_waiting, rrs_gotcc, rrs_gotaa, rrs_got1, rrs_finished} state = rrs_waiting;

  for (count = 0; state != rrs_finished; count++) {
    if (count == size) {
      fprintf(stderr, "Buffer size (%d) exceeded\n", size);
      exit(1);
    }
    while (read(fd, &buffer[count], 1) != 1) {
      if (errno != EAGAIN) {
	perror("Read2");
      }
    }
    switch (state) {
      case rrs_waiting:
	if (buffer[count] == 0xcc) state = rrs_gotcc;
	if (buffer[count] == 0xaa) state = rrs_gotaa;
	break;
      case rrs_gotcc:
	if (buffer[count] == 0x33) {
	  state = rrs_waiting;
	  count--;
	} else {
	  state = rrs_got1;
	}
	break;
      case rrs_gotaa:
	if (buffer[count] == 0x55) {
	  count--;
	}
	state = rrs_waiting;
	break;
      case rrs_got1:
	state = rrs_finished;
	break;
      default:
	printf("Illegal value in switch");
	exit(1);
    }
  }

  return count;
}

static void
unknown(unsigned char* buffer, int count)
{
}

static void
memmap(unsigned char* buffer, int count)
{
}

static void
regdump(unsigned char* buffer, int count)
{
}

static void
write_packet(int fd, int prefix, const unsigned char* data, int len)
{
  int b;
  unsigned int sum;
  unsigned char start_mark = 0xaa;
  unsigned char stop_mark = 0xcc;
  unsigned char start_esc_mark = 0x55;
  unsigned char stop_esc_mark = 0x33;
  unsigned char cdata;

  sum = 0x0;

  write(fd, &start_mark, 1);
  if (prefix != -1) {
    write(fd, &prefix, 1);
    sum += prefix;
  }

  for (b=0; b<len; b++) {
    if (b % 1000 == 0) {
      putchar('.');
      fflush(stdout);
    }
    write(fd, &data[b], 1);
    switch (data[b]) {
      case 0xaa:
	write(fd, &start_esc_mark, 1);
	break;
      case 0xcc:
	write(fd, &stop_esc_mark, 1);
	break;
    }
    sum += data[b];
  }

  write(fd, &stop_mark, 1);

  sum |= 0x8000;
  cdata = (sum >> 8) & 0xff;
  write(fd, &cdata, 1);
  cdata = sum & 0xff;
  write(fd, &cdata, 1);
}

static void
pack_word(unsigned char* buffer, int data)
{
  buffer[0] = (data >> 24) & 0xff;
  buffer[1] = (data >> 16) & 0xff;
  buffer[2] = (data >>  8) & 0xff;
  buffer[3] = (data >>  0) & 0xff;
}

void
sh3eva_write(const unsigned char* data, int len, unsigned long addr)
{
  unsigned char buffer[2560];
  int readlen;
  int a;
  int offset;
  int part_len;

  offset = 0;
  part_len = len;

  if (verbose) printf("Write header\n");
  buffer[0] = 0x06;
  pack_word(&buffer[1], addr + offset);
  pack_word(&buffer[5], part_len);
  write_packet(fd, -1, buffer, 1+4+4);

  for (a=0; a<1; a++) {
    readlen = read_result(fd, buffer, sizeof(buffer));
    if (verbose) dump_buffer(buffer, readlen);
  }

  if (verbose) printf("Write body\n");
  write_packet(fd, 0x0f, &data[offset], part_len);

  if (verbose) printf("Check result\n");
  for (a=0; a<1; a++) {
    readlen = read_result(fd, buffer, sizeof(buffer));
    if (verbose) dump_buffer(buffer, readlen);
  }
  
#if 0
    if (verbose) printf("Write crc req\n");
    buffer[0] = 0x0c;
    pack_word(&buffer[1], addr);
    pack_word(&buffer[5], len);
    pack_word(&buffer[9], 0);
    write_packet(fd, -1, buffer, 1+4+4+4);

    for (a=0; a<1; a++) {
      readlen = read_result(fd, buffer, sizeof(buffer));
      if (verbose) dump_buffer(buffer, readlen);
    }
#endif

#if 0
  if (verbose) printf("Write check\n");
  buffer[0] = 0x08;
  pack_word(&buffer[1], addr);
  pack_word(&buffer[5], len);
  write_packet(fd, -1, buffer, 1+4+4);

  for (a=0; a<2; a++) {
    readlen = read_result(fd, buffer, sizeof(buffer));
    if (verbose) dump_buffer(buffer, readlen);
  }
#endif
  putchar('\n');
}

void
sh3eva_run(unsigned long addr)
{
  unsigned char buffer[32];
  int readlen;
  int a;

  printf("Run (%08lx)\n", addr);

  buffer[0] = 0x04;
  buffer[1] = 0x1e;
  pack_word(&buffer[2], addr);
  write_packet(fd, -1, buffer, 2+4);

  readlen = read_result(fd, buffer, sizeof(buffer));
  if (verbose) dump_buffer(buffer, readlen);

  buffer[0] = 0x02;
  buffer[1] = 0x02;
  buffer[2] = 0x00;
  buffer[3] = 0x01;
  write_packet(fd, -1, buffer, 4);

  for (a=0; a<1; a++) {
    readlen = read_result(fd, buffer, sizeof(buffer));
    if (verbose) dump_buffer(buffer, readlen);
  }
  putchar('\n');
}

void
sh3eva_open(const char* filename)
{
  fd = local_open(filename);
}

void
sh3eva_connect(void)
{
  int a, b, c;
  int len;
  unsigned char buffer[256];

  printf("Connecting\n");
  for (a=0; a<SEND_DATA_LEN; a++) {
    /* Send the commands */
    b = a;
    if (verbose) printf("Send init string\n");
    write_packet(fd, -1, send_data[b].data, send_data[b].len);

    /* Read the results */
    if (verbose) printf("Read init string result\n");
    for (c = 0; c<send_data[b].results; c++) {
      len = read_result(fd, buffer, sizeof(buffer));
      if (verbose) dump_buffer(buffer, len);
      send_data[b].fn(buffer, c);
    }
  }
  putchar('\n');
}

void
sh3eva_terminal(void)
{
  int len;
  unsigned char buffer[256];

  printf("Terminal mode\n");
  while (1) {
    while ((len = read(fd, &buffer, sizeof(buffer))) <0) {
      if (errno != EAGAIN) {
	perror("Read2");
      }
    }
    if (verbose)
      dump_buffer(buffer, len);
    else
      write(1, buffer, len);
  }
}
